"""Creates a binary UpdateScript object containing the sensor graph and any config information."""

from __future__ import unicode_literals, absolute_import, print_function
import struct
from future.utils import viewitems
from iotile.sg.update import (AddNodeRecord, AddStreamerRecord, SetConfigRecord, SetConstantRecord,
                              SetGraphOnlineRecord, PersistGraphRecord, ClearDataRecord, ResetGraphRecord)
from iotile.core.hw.update.records import SetDeviceTagRecord
from iotile.core.hw.update import UpdateScript
from iotile.core.exceptions import ArgumentError


def format_script(sensor_graph):
    """Create a binary script containing this sensor graph.

    This function produces a repeatable script by applying a known sorting
    order to all constants and config variables when iterating over those
    dictionaries.

    Args:
        sensor_graph (SensorGraph): the sensor graph that we want to format

    Returns:
        bytearray: The binary script data.

    """

    records = []

    records.append(SetGraphOnlineRecord(False, address=8))
    records.append(ClearDataRecord(address=8))
    records.append(ResetGraphRecord(address=8))

    for node in sensor_graph.nodes:
        records.append(AddNodeRecord(str(node), address=8))

    for streamer in sensor_graph.streamers:
        records.append(AddStreamerRecord(streamer, address=8))

    for stream, value in sorted(viewitems(sensor_graph.constant_database), key=lambda x: x[0].encode()):
        records.append(SetConstantRecord(stream, value, address=8))

    records.append(PersistGraphRecord(address=8))

    for slot in sorted(sensor_graph.config_database, key=lambda x: x.encode()):
        for config_id in sorted(sensor_graph.config_database[slot]):
            config_type, value = sensor_graph.config_database[slot][config_id]
            byte_value = _convert_to_bytes(config_type, value)

            records.append(SetConfigRecord(slot, config_id, byte_value))

    # If we have an app tag and version set program them in
    app_tag = sensor_graph.metadata_database.get('app_tag')
    app_version = sensor_graph.metadata_database.get('app_version')

    if app_tag is not None:
        records.append(SetDeviceTagRecord(app_tag=app_tag, app_version=app_version))

    script = UpdateScript(records)
    return script.encode()


def _convert_to_bytes(type_name, value):
    """Convert a typed value to a binary array"""

    int_types = {'uint8_t': 'B', 'int8_t': 'b', 'uint16_t': 'H', 'int16_t': 'h', 'uint32_t': 'L', 'int32_t': 'l'}

    type_name = type_name.lower()

    if type_name not in int_types and type_name not in ['string', 'binary']:
        raise ArgumentError('Type must be a known integer type, integer type array, string', known_integers=int_types.keys(), actual_type=type_name)

    if type_name == 'string':
        #value should be passed as a string
        bytevalue = bytearray(value)
    elif type_name == 'binary':
        bytevalue = bytearray(value)
    else:
        bytevalue = struct.pack("<%s" % int_types[type_name], value)

    return bytevalue



"""Parsing routines for understanding binary streamer descriptors.

Binary streamer descriptors are the internal binary representation of
a streamer stored in an IOTile device's embedded firmware.
"""

from __future__ import (unicode_literals, absolute_import, print_function)
import struct
from typedargs.exceptions import ArgumentError
from .slot import SlotIdentifier
from .stream import DataStreamSelector
from .streamer import DataStreamer


def parse_binary_descriptor(bindata):
    """Convert a binary streamer descriptor into a string descriptor.

    Binary streamer descriptors are 20-byte binary structures that encode all
    information needed to create a streamer.  They are used to communicate
    that information to an embedded device in an efficent format.  This
    function exists to turn such a compressed streamer description back into
    an understandable string.

    Args:
        bindata (bytes): The binary streamer descriptor that we want to
            understand.

    Returns:
        DataStreamer: A DataStreamer object representing the streamer.  You
            can get a useful human readable string by calling str() on the
            return value.
    """

    if len(bindata) != 14:
        raise ArgumentError("Invalid length of binary data in streamer descriptor", length=len(bindata), expected=14, data=bindata)

    dest_tile, stream_id, trigger, format_code, type_code = struct.unpack("<8sHBBBx", bindata)

    dest_id = SlotIdentifier.FromEncoded(dest_tile)
    selector = DataStreamSelector.FromEncoded(stream_id)

    format_name = DataStreamer.KnownFormatCodes.get(format_code)
    type_name = DataStreamer.KnownTypeCodes.get(type_code)

    if format_name is None:
        raise ArgumentError("Unknown format code", code=format_code, known_code=DataStreamer.KnownFormatCodes)
    if type_name is None:
        raise ArgumentError("Unknown type code", code=type_code, known_codes=DataStreamer.KnownTypeCodes)

    with_other = None
    if trigger & (1 << 7):
        auto = False
        with_other = trigger & ((1 << 7) - 1)
    elif trigger == 0:
        auto = False
    elif trigger == 1:
        auto = True
    else:
        raise ArgumentError("Unknown trigger type for streamer", trigger_code=trigger)

    return DataStreamer(selector, dest_id, format_name, auto, type_name, with_other=with_other)


def create_binary_descriptor(streamer):
    """Create a packed binary descriptor of a DataStreamer object.

    Args:
        streamer (DataStreamer): The streamer to create a packed descriptor for

    Returns:
        bytes: A packed 14-byte streamer descriptor.
    """

    trigger = 0
    if streamer.automatic:
        trigger = 1
    elif streamer.with_other is not None:
        trigger = (1 << 7) | streamer.with_other

    return struct.pack("<8sHBBBx", streamer.dest.encode(), streamer.selector.encode(), trigger, streamer.KnownFormats[streamer.format], streamer.KnownTypes[streamer.report_type])



"""An UpdateScript record that sets a config variable."""

from __future__ import unicode_literals, absolute_import, print_function
import struct
from binascii import hexlify
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.update.record import MatchQuality, UpdateRecord
from iotile.core.hw.update.records import SendErrorCheckingRPCRecord
from iotile.sg import SlotIdentifier


#pylint: disable=abstract-method;Method encode_contents is not overriden because we directly override encode
@python_2_unicode_compatible
class SetConfigRecord(UpdateRecord):
    """Set a config value on a specific tile.

    Args:
        tile (SlotIdentifier): The target tile
        config_id (int): The id of the config variable we want to set
        data (bytearray): The raw data that we wish to load into the
            config variable.
    """

    RecordType = 4

    BEGIN_CONFIG_RPC = 0x2A07
    PUSH_CONFIG_RPC = 0x2A08
    END_CONFIG_RPC = 0x2A09

    def __init__(self, tile, config_id, data):
        self.target = tile
        self.config_id = config_id
        self.data = data

    def encode(self):
        """Encode this record into binary, suitable for embedded into an update script.

        This function will create multiple records that correspond to the actual
        underlying rpcs that SetConfigRecord turns into.

        Returns:
            bytearary: The binary version of the record that could be parsed via
                a call to UpdateRecord.FromBinary()
        """

        begin_payload = struct.pack("<H8s", self.config_id, self.target.encode())
        start_record = SendErrorCheckingRPCRecord(8, self.BEGIN_CONFIG_RPC, begin_payload, 4)
        end_record = SendErrorCheckingRPCRecord(8, self.END_CONFIG_RPC, bytearray(), 4)
        push_records = []

        for i in range(0, len(self.data), 20):
            chunk = self.data[i:i+20]
            push_record = SendErrorCheckingRPCRecord(8, self.PUSH_CONFIG_RPC, chunk, 4)
            push_records.append(push_record)

        out_blob = bytearray()
        out_blob += start_record.encode()

        for push_record in push_records:
            out_blob += push_record.encode()

        out_blob += end_record.encode()
        return out_blob

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        return cls.RecordType

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this functon should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count == 1:
            cmd, address, _resp_length, _payload = SendErrorCheckingRPCRecord._parse_rpc_info(record_data)

            if cmd == cls.BEGIN_CONFIG_RPC and address == 8:
                return MatchQuality.DeferMatch

            return MatchQuality.NoMatch

        # To see if this is a set_config variable record set, we need to decode all of
        # the records and make sure each is an error checking rpc with the right rpc id
        try:
            rpcs = SendErrorCheckingRPCRecord.parse_multiple_rpcs(record_data)

            push_commands = rpcs[1:-1]

            for cmd in push_commands:
                cmd_id, addr = cmd[:2]
                if cmd_id != cls.PUSH_CONFIG_RPC or addr != 8:
                    return MatchQuality.NoMatch

            last_cmd, last_addr = rpcs[-1][:2]
            if last_cmd == cls.END_CONFIG_RPC and last_addr == 8:
                return MatchQuality.PerfectMatch
        except ArgumentError:
            return MatchQuality.NoMatch

        return MatchQuality.DeferMatch

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This is a multi-action record that matches a pattern of error checking
        RPC calls:
        begin config
        push config data
        <possibly multiple>
        end config

        Args:
            record_data (bytearray): The raw record data that we wish to parse.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            SetConfigRecord: The decoded tile records.
        """

        rpcs = SendErrorCheckingRPCRecord.parse_multiple_rpcs(record_data)

        start_rpc = rpcs[0]
        push_rpcs = rpcs[1:-1]

        try:
            config_id, raw_target = struct.unpack("<H8s", start_rpc.payload)
            target = SlotIdentifier.FromEncoded(raw_target)
        except ValueError:
            raise ArgumentError("Could not parse payload on begin config rpc", payload=start_rpc.payload)

        payload = bytearray()
        for rpc in push_rpcs:
            payload += rpc.payload

        return SetConfigRecord(target, config_id, payload)

    def __str__(self):
        return "Set config variable 0x%X on %s to 'hex:%s'" % (self.config_id, self.target, hexlify(self.data))
